/***************************************************************************
 *            hybrid_system_exercise.cpp
 *
 *  Copyright  2022  Luca Geretti
 *
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ariadne.hpp>
#include <ariadne_main.hpp>

using namespace Ariadne;

void ariadne_main()
{
    // Pool tank (z) and reserve tank (d)
    HybridAutomaton tanks("tanks");
    RealVariable z("z"), d("d"), a("a");
    RealConstant alpha("alpha",0.1_dec), beta("beta",0.4_dec), gamma("gamma",0.1_dec);
    RealVariable w("w"), r("r"), e("e");

    tanks.new_mode({dot(z)=a*beta*d-alpha*z+r*gamma,dot(d)=-a*beta*d+alpha*z});

    // Weather (-1<=r<=1, with r=1 totally raining and r=-1 no rain)
    StringVariable weather_name("weather");
    HybridAutomaton weather(weather_name.name());
    RealConstant c("c",0.1_dec), thr_m("thr_m",0.2_dec), thr_M("thr_M",2.2_dec);
    RealVariable clk("clk");
    RealConstant sunny_time("sunny_time",10);
    RealConstant stormy_time("stormy_time",10);
    DiscreteLocation sunny(weather_name|"sunny"), stormy(weather_name|"stormy"), decreasing_rain(weather_name|"decreasing_rain"),
                     increasing_rain(weather_name|"increasing_rain");
    weather.new_mode(decreasing_rain,{let(r)=w-thr_m-1,let(clk)=0},{dot(w)=-c*(w+w*w)});
    weather.new_mode(increasing_rain,{let(r)=w-thr_m-1,let(clk)=0},{dot(w)=c*(w+w*w)});
    weather.new_mode(sunny,{let(r)=-1,let(w)=thr_m},{dot(clk)=1});
    weather.new_mode(stormy,{let(r)=1,let(w)=thr_M},{dot(clk)=1});

    DiscreteEvent clearsup("clearsup");
    DiscreteEvent clouds("clouds");
    DiscreteEvent cleared("cleared");
    DiscreteEvent pours("pours");
    weather.new_transition(decreasing_rain,cleared,sunny,{next(clk)=0},w<=thr_m,EventKind::URGENT);
    weather.new_transition(increasing_rain,pours,stormy,{next(clk)=0},w>=thr_M,EventKind::URGENT);
    weather.new_transition(sunny,clouds,increasing_rain,{next(w)=w},clk>=sunny_time,EventKind::URGENT);
    weather.new_transition(stormy,clearsup,decreasing_rain,{next(w)=w},clk>=stormy_time,EventKind::URGENT);

    // Controller
    RealConstant zmin("zmin",1.0_dec);
    RealConstant zmax("zmax",1.9_dec);

    DiscreteEvent open("open");
    DiscreteEvent close("close");

    StringVariable controller_name("controller");
    HybridAutomaton controller(controller_name.name());

    DiscreteLocation valve_opened(controller_name|"valve_opened");
    DiscreteLocation valve_closed(controller_name|"valve_closed");

    controller.new_mode(valve_opened,{let(a)=1});
    controller.new_mode(valve_closed,{let(a)=0});

    controller.new_transition(valve_closed,open,valve_opened,z<=zmin,EventKind::URGENT);
    controller.new_transition(valve_opened,close,valve_closed,z>=zmax,EventKind::URGENT);

    CompositeHybridAutomaton composition("system",{tanks,weather,controller});

    // Analysis

    HybridBoundedConstraintSet initial_set({weather_name|"decreasing_rain",controller_name|"valve_opened"},{w==0.5_dec,z==1.8_dec,d==0.5_dec});
    Real continuous_time = 300.0_dec;
    HybridTime final_time(continuous_time,15);

    HybridSimulator simulator(composition);
    simulator.configuration().set_step_size(0.01);
    ARIADNE_LOG_PRINTLN_VAR(simulator.configuration());

    ARIADNE_LOG_PRINTLN("Computing simulation trajectory...");
    auto orbit = simulator.orbit(initial_set,final_time);

    ARIADNE_LOG_PRINTLN("Plotting simulation trajectory...");
    plot("simulation_t-z",Axes2d(0<=TimeVariable()<=continuous_time,-10<=z<=10),orbit);
    plot("simulation_t-d",Axes2d(0<=TimeVariable()<=continuous_time,0<=d<=10),orbit);
    plot("simulation_z-d",Axes2d(0<=z<=2,0<=d<=10),orbit);
    plot("simulation_r-w",Axes2d(0<=r<=1,0<=w<=1),orbit);
    plot("simulation_t-r",Axes2d(0<=TimeVariable()<=continuous_time,-1<=r<=1),orbit);
    ARIADNE_LOG_PRINTLN("Done computing and plotting simulation trajectory!");
}
